module stream_file;
import stream;

version(Win32)
{
	private import windows;
	private extern(Windows)
	{
		HANDLE GetStdHandle(uint);
		int SetEndOfFile(HANDLE);
		int CopyFileA(char*, char*, int);
		int MoveFile(char*, char*);
		int DeleteFileA(char*);
	}
	enum Mode
	{
		r  = GENERIC_READ | OPEN_EXISTING,			// reading
		w  = GENERIC_WRITE| OPEN_ALWAYS,			// writing
		rw = GENERIC_READ | GENERIC_WRITE | OPEN_ALWAYS,	// reading and writing
		wr = GENERIC_READ | GENERIC_WRITE | OPEN_ALWAYS,	// same as rw
		c  = GENERIC_WRITE| CREATE_ALWAYS,			// create the file (writing)
		rc = GENERIC_READ | GENERIC_WRITE | CREATE_ALWAYS,	// create the file (reading and writing)
		cr = GENERIC_READ | GENERIC_WRITE | CREATE_ALWAYS,	// same as rc
	}
}
version(linux)
{
	private import linux;
	enum Mode
	{
		r  = O_RDONLY,				// reading
		w  = O_WRONLY | O_CREAT,		// writing
		rw = O_RDWR   | O_CREAT,		// reading and writing
		wr = O_RDWR   | O_CREAT,		// same as rw
		c  = O_WRONLY | O_CREAT | O_TRUNC,	// create the file (writing)
		rc = O_RDWR   | O_CREAT | O_TRUNC,	// create the file (reading and writing)
		cr = O_RDWR   | O_CREAT | O_TRUNC	// same as rc
	}
}

alias FileStream File;
class FileStream : Stream
{
	// This stream is bound to a file.
	// Every operation on the stream will affect
	// the file the stream is bound to.

	// A file name and an access mode can be passed
	// to the constructor. If no access mode is specified
	// (that is only the file name is passed) then the file
	// is opened for reading and writing.

	this()				{                       }
	this(char[] filename)		{ open(filename);       }
	this(char[] filename, Mode mode){ open(filename, mode); }

	this(HANDLE handle)
	in
	{
	    version(Win32)
	    {
		assert(handle != null);
	    }
	    version(linux)
	    {
		assert(handle != -1);
	    }
	}
	body
	{
		h = handle;
	}

	~this()
	{
		if(h) close();
	}

	import string;
	void open(char[] filename)
	{
		open(filename, Mode.rw);
	}

	void open(char[] filename, Mode mode)
	in
	{
		assert(filename);
		assert(filename.length);
	}
	body
	{
	    version(Win32)
	    {
		h = CreateFileA
		(
			toStringz(filename),
			mode & 0xFFFF0000,
			FILE_SHARE_READ,
			null,
			mode & 0x0000FFFF,
			FILE_ATTRIBUTE_NORMAL,
			null
		);
		if(!h) throw new StreamError("FileStream: unable to open file " ~ filename);
	    }
	    version(linux)
	    {
		h = linux.open(toStringz(filename), mode, 0x01B0);
		if(h == -1) throw new StreamError("FileStream: unable to open file " ~ filename);
	    }
	}

	void close()
	{
		// WHAT IF STDIN OR STDOUT ARE CLOSED?
		// SHOULD THIS BE PREVENTED?
	    version(Win32)
	    {
		CloseHandle(h);
		h = null;
	    }
	    version(linux)
	    {
		linux.close(h);
		h = -1;
	    }
	}

	void seek(int offset, Rel rel)
	in
	{
		if(rel == Rel.begin) assert(offset >= 0);
		if(rel == Rel.end)   assert(offset <= 0);
	}
	body
	{
	    version(Win32)
	    {
		if(SetFilePointer(h, offset, null, rel) == ~0)
			throw new SeekError("FileStream: unable to move file pointer");
	    }
	    version(linux)
	    {
		if(linux.lseek(h, offset, rel) == ~0)
			throw new SeekError("FileStream: unable to move file pointer");
	    }
	}

	ulong position()
	{
	    version(Win32)
	    {
		return SetFilePointer(h, 0, null, Rel.cur);
	    }
	    version(linux)
	    {
		return linux.lseek(h, 0, Rel.cur);
	    }
	}

	ulong length()
	{
	    version(Win32)
	    {
		ulong len = 0;
		len += GetFileSize(h, cast(uint*)&len + 1);
		if(cast(uint) len == ~0)
			throw new StreamError("FileStream: unable to get file length");
		return len;
	    }
	    version(linux)
	    {
		assert(false); // TO BE IMPLEMENTED
	    }
	}

	void truncate()
	{
	    version(Win32)
	    {
		if(!SetEndOfFile(h))
			throw new WriteError("FileStream: unable to truncate file");
	    }
	    version(linux)
	    {
		assert(false); // TO BE IMPLEMENTED
	    }
	}

	// Some static functions to copy, move and delete files.
	// For example: File.copy('readme.txt', '..\readme.txt');
	// would work as "copy readme.txt ..".

	static void copy(char[] filename, char[] destfile)
	{
	    version(Win32)
	    {
		if(!CopyFileA(toStringz(filename), toStringz(destfile), true))
			throw new StreamError("FileStream: unable to copy file " ~ filename);
	    }
	    version(linux)
	    {
		assert(false); // TO BE IMPLEMENTED
	    }
	}

	static void move(char[] filename, char[] destfile)
	{
	    version(Win32)
	    {
		if(!MoveFile(toStringz(filename), toStringz(destfile)))
			throw new StreamError("FileStream: unable to move file " ~ filename);
	    }
	    version(linux)
	    {
		assert(false); // TO BE IMPLEMENTED
	    }
	}

	static void erase(char[] filename)
	{
	    version(Win32)
	    {
		if(!DeleteFileA(filename))
			throw new StreamError("FileStream: unable to erase file " ~ filename);
	    }
	    version(linux)
	    {
		assert(false); // TO BE IMPLEMENTED
	    }
	}

		protected:

	version(Win32) HANDLE h = null;
	version(linux) HANDLE h = -1;

	void readBlock(void* buf, uint buflen)
	{
		assert(buf);
		uint readb;

	    version(Win32)
	    {
		if(!ReadFile(h, buf, buflen, &readb, null))
			goto error;
	    }
	    version(linux)
	    {
		readb = linux.read(h, buf, buflen);
	    }

		if(readb == buflen) return;
		error: throw new ReadError("FileStream: unable to read from file");
	}

	void writeBlock(void* buf, uint buflen)
	{
		assert(buf);
		uint writtenb;

	    version(Win32)
	    {
		if(!WriteFile(h, buf, buflen, &writtenb, null))
			goto error;
	    }
	    version(linux)
	    {
		writtenb = linux.write(h, buf, buflen);
	    }

		if(writtenb == buflen) return;
		error: throw new WriteError("FileStream: unable to write to file");
	}
}

// stdin, stdout and stderr streams.
// Used to read input from the keyboard and output to the shell.
// You can not seek in these streams.

FileStream stdin, stdout, stderr;

version(Win32) static this()
{
	stdin  = new FileStream(GetStdHandle(-10));
	stdout = new FileStream(GetStdHandle(-11));
	stderr = new FileStream(GetStdHandle(-12));
}
version(linux) static this()
{
	stdin  = new FileStream(0);
	stdout = new FileStream(1);
	stderr = new FileStream(2);
}