module stream_slice;
import stream;

class SliceStream : Stream
{
	// These stream is bound to a slice
	// of another stream. It is lower bounded
	// and can also be higher bounded.

	// Seeking before the lower bound or
	// after the higher bound will not raise
	// an exception, but the stream pointer
	// will be set to the bound itself.

	this(Stream base, ulong low)
	in
	{
		assert(base);
		assert(low <= base.length());
	}
	body
	{
		this.base = base;
		this.low = low;
		highbounded = false;
		if(base.position() < low)
			base.seek(low, Rel.begin);
	}

	this(Stream _base, ulong _low, ulong _high)
	in
	{
		assert(base);
		assert(low <= high);
		assert(high <= base.length());
	}
	body
	{
		this.base = base;
		this.low = low;
		this.high = high;
		highbounded = true;
		if(base.position() < low)
			base.seek(low, Rel.begin);
		else if(base.position() > high)
			base.seek(high, Rel.begin);
	}

	void seek(int offset, Rel rel)
	in
	{
		if(rel == Rel.begin) assert(offset >= 0);
		if(rel == Rel.end)   assert(offset <= 0);
	}
	body
	{
		long where;
		switch(rel)
		{
			case Rel.begin:	where = low + offset; break;
			case Rel.cur:	where = base.position() + offset; break;
			case Rel.end:	where = (highbounded ? high : base.length()) + offset; break;
		}

		if(where < low) where = low;
		if(highbounded && where > high) where = high;
		base.seek(where, Rel.begin);
	}

	ulong position()
	{
		return base.position() - low;
	}

	ulong length()
	{
		return (highbounded ? high : base.length()) - low;
	}

	void truncate()
	{
		base.truncate();
		highbounded = false;
	}

		protected:

	Stream base;
	ulong low;
	ulong high;
	bit highbounded;

	void readBlock(void* buf, uint buflen)
	{
		if(highbounded && base.position() + buflen > high)
			throw new ReadError("SliceStream: reading over higher bound");

		base.readBlock(buf, buflen);
	}

	void writeBlock(void* buf, uint buflen)
	{
		if(highbounded && base.position() + buflen > high)
			throw new ReadError("SliceStream: writing over higher bound");

		base.writeBlock(buf, buflen);
	}
}